from abc import ABC, abstractmethod
import logging
from enum import auto
from strenum import StrEnum


log = logging.getLogger("Output")


class OutputType(StrEnum):
    SCREEN = auto()
    MQTT = auto()


class AbstractOutput(ABC):
    def __init__(self, formatter):
        self.formatter = formatter

    @abstractmethod
    def output(self, data):
        pass



# !/usr/bin/python3
"""main powermon code"""

import logging
from argparse import ArgumentParser

# from collections import deque
from datetime import date, timedelta  # noqa: F401

# from time import sleep, time

import yaml


from mppsolar.version import __version__  # noqa: F401
from powermon.libs.daemon import Daemon
from powermon.libs.mqttbroker import MqttBroker

from powermon.libs.schedule import Schedule
from powermon.libs.device import Device
from powermon.libs.apicoordinator import ApiCoordinator

# from powermon.ports import getPortFromConfig


# Set-up logger
log = logging.getLogger("")


def read_yaml_file(yaml_file=None):
    """function to read a yaml file and return dict"""
    _yaml = {}
    if yaml_file is not None:
        try:
            with open(yaml_file, "r", encoding="utf-8") as stream:
                _yaml = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            log.error("Error processing yaml file: %s", exc)
        except FileNotFoundError as exc:
            log.error("Error opening yaml file: %s", exc)
    return _yaml


def process_command_line_overrides(args):
    """override config with command line options"""
    _config = {}
    if args.once:
        _config["loop"] = "once"
    if args.info:
        _config["debuglevel"] = logging.INFO
    if args.debug:
        _config["debuglevel"] = logging.DEBUG
    return _config


def main():
    """main entry point for powermon command"""
    description = f"Power Device Monitoring Utility, version: {__version__}"
    parser = ArgumentParser(description=description)

    parser.add_argument(
        "-C",
        "--configFile",
        nargs="?",
        type=str,
        help="Full location of config file",
        const="./powermon.yaml",
        default=None,
    )
    parser.add_argument(
        "-v", "--version", action="store_true", help="Display the version"
    )
    parser.add_argument(
        "-d",
        "--dumpConfig",
        action="store_true",
        help="Print the config based on options supplied (including the defaults, config file and any overrides)",
    )
    parser.add_argument(
        "-1",
        "--once",
        action="store_true",
        help="Only loop through config once",
    )
    parser.add_argument(
        "-D",
        "--debug",
        action="store_true",
        help="Enable Debug and above (i.e. all) messages",
    )
    parser.add_argument(
        "-I", "--info", action="store_true", help="Enable Info and above level messages"
    )
    parser.add_argument(
        "--adhoc",
        type=str,
        default=None,
        help="Send adhoc command to mqtt adhoc command queue - needs config file specified and populated",
    )

    args = parser.parse_args()
    # prog_name = parser.prog

    # Temporarily set debug level based on command line options
    log.setLevel(logging.WARNING)
    if args.info:
        log.setLevel(logging.INFO)
    if args.debug:
        log.setLevel(logging.DEBUG)

    # Display version if asked
    log.info(description)
    if args.version:
        print(description)
        return None

    # Build configuration from config file and command line overrides
    log.info("Using config file: %s", args.configFile)
    # build config with details from config file
    config = read_yaml_file(args.configFile)

    # build config - override with any command line arguments
    config.update(process_command_line_overrides(args))

    # logging (DEBUG, INFO, WARNING, ERROR, CRITICAL)
    log.setLevel(config.get("debuglevel", logging.WARNING))

    # if generateConfigFile is true then print config out
    if args.dumpConfig:
        print("# yaml config for powermon")
        print("# default location ./powermon.yaml")
        print(yaml.dump(config))
        return
    # debug config
    log.info("config: %s", config)

    # build device object (required)
    device = Device(config=config.get("device", None))
    log.debug("device: %s", device)

    # build mqtt broker object (optional)
    mqtt_broker = MqttBroker(config=config.get("mqttbroker", {}))
    log.debug("mqtt_broker: %s", mqtt_broker)

    # build the daemon object (optional)
    daemon = Daemon(config=config)
    log.debug("daemon: %s", daemon)

    # Get scheduled commands
    scheduling_config = config.get("scheduling", None)
    log.debug("scheduling_config: %s", scheduling_config)
    schedule = Schedule.parseScheduleConfig(scheduling_config, device, mqtt_broker)
    log.debug("schedule: %s", schedule)

    # setup api coordinator
    api_coordinator = ApiCoordinator(
        config=config.get("api", None),
        device=device,
        mqtt_broker=mqtt_broker,
        schedule=schedule,
    )
    # TODO: run in the schedule loop

    # initialize daemon
    daemon.initialize()

    # Main working loop
    keep_looping = True
    try:
        schedule.beforeLoop()
        while keep_looping:
            # tell the daemon we're still working
            daemon.watchdog()
            keep_looping = schedule.runLoop()
            api_coordinator.run()
    except KeyboardInterrupt:
        print("KeyboardInterrupt")
    except Exception as general_exception:
        print(general_exception)
    finally:
        # Disconnect port
        # port.disconnect()
        # Disconnect mqtt
        mqtt_broker.stop()
        # Notify the daemon
        daemon.stop()



"""device.py"""
import logging

from dto.deviceDTO import DeviceDTO
from powermon.ports import getPortFromConfig

# Set-up logger
log = logging.getLogger("Device")


class ConfigError(Exception):
    """Exception for invaild configurations"""


class Device:
    """
    A device is a port with a protocol
    also contains the name, model and id of the device
    """

    def __init__(self, config):
        if not config:
            log.warning("No device definition in config. Check configFile?")
            config = {}
            # sys.exit(1)
        self.name = config.get("name", "mppsolar")
        self.identifier = config.get("identifier", "mppsolar")
        self.model = config.get("model", "mppsolar")
        self.manufacturer = config.get("manufacturer", "mppsolar")

        self.port = getPortFromConfig(config.get("port"))
        # error out if unable to configure port
        if not self.port:
            log.error("Invalid port config '%s' found", config)
            raise ConfigError(f"Invalid port config '{config}' found")

    def toDTO(self) -> DeviceDTO:
        dto = DeviceDTO(
            name=self.name,
            identifier=self.identifier,
            model=self.model,
            manufacturer=self.manufacturer,
            port=self.port.toDTO(),
        )
        return dto



import logging
from enum import Enum, auto
from time import time


# Set-up logger
log = logging.getLogger("daemon")


class dummyNotification(Enum):
    READY = auto()
    STATUS = auto()
    STOPPING = auto()
    WATCHDOG = auto()


class Daemon:
    def __str__(self):
        return f"Daemon name: {self.type}"

    def __init__(self, config={}):

        daemon_config = config.get("daemon")
        if daemon_config is None:
            self.type = None
            self.keepalive = 60
        if daemon_config is not None:
            self.type = daemon_config.get("type", None)
            self.keepalive = daemon_config.get("keepalive", 60)

        log.info(f"got daemon type: {self.type}, keepalive: {self.keepalive}")

        if self.type == "systemd":
            try:
                from cysystemd.daemon import Notification, notify
                from cysystemd import journal

                self._notify = notify
                self._journal = journal.write
                self._Notification = Notification
            except ModuleNotFoundError as e:
                print(
                    f"error: {e}, try 'pip install cysystemd' (which may need 'apt install build-essential libsystemd-dev'), see https://pypi.org/project/cysystemd/ for further info"
                )
                exit(1)
        else:
            self._notify = self._dummyNotify
            self._journal = self._dummyNotify
            self._Notification = dummyNotification
        self.notify(f"got daemon type: {self.type}, keepalive: {self.keepalive}")

    def initialize(self, *args, **kwargs):
        # Send READY=1
        self._notify(self._Notification.READY)
        self._lastNotify = time()

    def watchdog(self, *args, **kwargs):
        elapsed = time() - self._lastNotify
        if (elapsed) > self.keepalive:
            self._notify(self._Notification.WATCHDOG)
            self._lastNotify = time()
            self._journal(f"Daemon notify at {self._lastNotify}")

    def notify(self, *args, **kwargs):
        # Send status
        if args:
            status = args[0]
        else:
            status = "OK"
        self._notify(self._Notification.STATUS, status)

    def stop(self, *args, **kwargs):
        # Send stopping
        self._notify(self._Notification.STOPPING)

    def log(self, *args, **kwargs):
        # Print log message
        if args:
            self._journal(args[0])

    def _dummyNotify(self, *args, **kwargs):
        # Print log message
        # if args:
        #     print(args[0])
        pass



import logging
from enum import auto
from time import sleep, time

from strenum import LowercaseStrEnum

from dto.commandDTO import CommandDTO
from dto.commandScheduleDTO import CommandScheduleDTO
from dto.scheduleDTO import ScheduleDTO
from powermon.outputs import getOutputFromConfig

from .device import Device

log = logging.getLogger("Schedule")


class Command:
    def __init__(self, command, commandType, outputs, port):
        self.command = command
        self.commandType = commandType
        self.outputs = outputs
        self.port = port

    def __str__(self):
        return f"Command: {self.command}, CommandType: {self.commandType}, Outputs: {self.outputs}"

    def toDTO(self):
        dto = CommandDTO(
            command=self.command,
            commandType=self.commandType,
        )
        return dto

    def run(self):
        log.debug(f"Running command: {self.command}")
        results = self.port.process_command(command=self.command)
        for output in self.outputs:
            log.debug(f"Output: {output}")
            output.output(data=results)


class Schedule:
    def __init__(self, scheduledCommands, loopDuration, mqtt_broker, device: Device):
        self.scheduledCommands = scheduledCommands
        self.loopDuration = loopDuration
        self.inDelay = False
        self.mqtt_broker = mqtt_broker
        self.device = device

        if loopDuration == "once":
            self.delayRemaining = 0
        else:
            self.delayRemaining = loopDuration

    def __str__(self):
        return f"Schedule: {self.scheduledCommands}, {self.loopDuration}"

    def toDTO(self) -> ScheduleDTO:
        commandSchedules = []
        for scheduledCommand in self.scheduledCommands:
            commandSchedules.append(scheduledCommand.toDTO())
        dto = ScheduleDTO(
            loopDuration=self.loopDuration,
            device=self.device.toDTO(),
            schedulesCommands=commandSchedules,
        )
        return dto

    def addOneTimeCommandFromConfig(self, commandConfig):
        command = self.parseCommandConfig(commandConfig, self.mqtt_broker, self.device)
        self.scheduledCommands.append(OneTimeCommandSchedule({command}))

    # The hook for the port to connect before the main loops starts
    def beforeLoop(self):
        self.device.port.connect()

    def runLoop(self) -> bool:
        start_time = time()
        if self.inDelay is False:
            for scheduledCommand in self.scheduledCommands:
                if scheduledCommand.is_due():
                    for command in scheduledCommand.commands:
                        command.run()

        # Small pause to ....
        sleep(0.5)
        elapsed_time = time() - start_time
        self.delayRemaining = self.delayRemaining - elapsed_time
        if self.delayRemaining > 0 and not self.inDelay:
            log.debug(
                f"delaying for {self.loopDuration}sec, delayRemaining: {self.delayRemaining}"
            )
            self.inDelay = True
        if self.delayRemaining < 0:
            log.debug(f"Next loop: {self.loopDuration}, {self.delayRemaining}")
            self.inDelay = False
            self.delayRemaining = self.loopDuration

        if self.loopDuration == "once":
            log.debug("loopDuration is once, returning False")
            return False
        else:
            return True

    # TODO: this should follow the same pattern as the other parsers
    @classmethod
    def parseCommandConfig(cls, command, mqtt_broker, device) -> Command:
        _command = command["command"]
        _commandType = command["type"]
        _outputs = []
        for outputConfig in command["outputs"]:
            logging.debug(f"command: {command}")
            _output = getOutputFromConfig(outputConfig, device, mqtt_broker)
            logging.debug(f"output: {_output}")
            _outputs.append(_output)

        return Command(_command, _commandType, _outputs, device.port)

    # TODO: this should follow the same pattern as the other parsers
    @classmethod
    def parseScheduleConfig(cls, config, device, mqtt_broker):
        logging.debug("parseScheduleConfig: config: %s, device: %s, mqtt_broker: %s" % (config, device, mqtt_broker))
        if not config:
            log.info("no schedule config supplied, defaulting")
            config = {"loop": "once"}
        _loopDuration = config.get("loop")

        _schedules = []
        for schedule in config["schedules"]:
            _scheduleType = schedule["type"]
            if _scheduleType == CommandScheduleType.LOOP:
                _loopCount = schedule["loopCount"]
            elif _scheduleType == CommandScheduleType.TIME:
                _runTime = schedule["runTime"]

            _commands = []
            for command in schedule["commands"]:
                _commands.append(cls.parseCommandConfig(command, mqtt_broker, device))

            if _scheduleType == CommandScheduleType.LOOP:
                _schedules.append(LoopCommandSchedule(_loopCount, _commands))
            elif _scheduleType == CommandScheduleType.ONCE:
                _schedules.append(OneTimeCommandSchedule(_commands))
            else:
                raise KeyError(f"Undefined schedule type: {_scheduleType}")

        schedule = Schedule(_schedules, _loopDuration, mqtt_broker, device)
        return schedule


class CommandScheduleType(LowercaseStrEnum):
    LOOP = auto()
    TIME = auto()
    ONCE = auto()


class InformalScheduledCommandInterface:
    def is_due(self):
        raise NotImplementedError


class LoopCommandSchedule(InformalScheduledCommandInterface):
    def __init__(self, loopCount, commands: list):
        self.scheduleType = CommandScheduleType.LOOP
        self.loopCount = loopCount
        self.commands = commands

        self.currentLoopCount = (
            loopCount  # Set to loopCount so that the first loop will run
        )

    def __str__(self):
        return f"ScheduleType: {self.scheduleType}, LoopCount: {self.loopCount}, Commands: {self.commands}"

    def toDTO(self):
        commandDTOs = []
        for command in self.commands:
            commandDTOs.append(command.toDTO())
        dto = CommandScheduleDTO(type=self.scheduleType, commands=commandDTOs)
        return dto

    def is_due(self):
        if self.currentLoopCount < self.loopCount:
            self.currentLoopCount += 1
            return False
        else:
            self.currentLoopCount = 1
            return True


class OneTimeCommandSchedule(InformalScheduledCommandInterface):
    def __init__(self, _commands):
        self.scheduleType = CommandScheduleType.ONCE
        self.commands = _commands
        self.has_run = False

    def __str__(self):
        return f"ScheduleType: {self.scheduleType}, Commands: {self.commands}"

    def toDTO(self):
        commandDTOs = []
        for command in self.commands:
            commandDTOs.append(command.toDTO())
        dto = CommandScheduleDTO(scheduleType=self.scheduleType, commands=commandDTOs)
        return dto

    def is_due(self):
        if self.has_run:
            log.debug("One time command has already run")
            return False
        else:
            log.debug("One time command is due to run")
            self.has_run = True
            return True



from abc import ABC, abstractmethod
import logging
import re
from enum import auto
from strenum import StrEnum

# from time import sleep
log = logging.getLogger("Formatter")


class FormatterType(StrEnum):
    HASS = auto()
    HTMLTABLE = auto()
    RAW = auto()
    SIMPLE = auto()
    TABLE = auto()
    TOPICS = auto()


class AbstractFormat(ABC):
    def __init__(self, formatConfig):
        self.remove_spaces = formatConfig.get("remove_spaces", True)
        self.keep_case = formatConfig.get("keep_case", False)

        self._keyFilter = None
        _keyFilterString = formatConfig.get("filter", None)
        if _keyFilterString is not None:
            self._keyFilter = re.compile(_keyFilterString)

        self._keyExclusionfilter = None
        _keyExclusionFilterString = formatConfig.get("excl_filter", None)
        if _keyExclusionFilterString is not None:
            self._keyExclusionfilter = re.compile(_keyExclusionFilterString)

    @abstractmethod
    def format(self, data):
        pass

    # Override this if the format sends multiple messages
    def sendsMultipleMessages(self) -> bool:
        return False

    def formatAndFilterData(self, data):
        # TODO: should we make data a proper object so it's easy to get the data we want?
        # remove raw response
        if "raw_response" in data:
            data.pop("raw_response")
        # remove command details
        if "_command" in data:
            data.pop("_command")
        if "_command_description" in data:
            data.pop("_command_description")

        displayData = {}
        for key in data:
            _values = data[key]
            formattedKey = self.formatKey(key)
            if self.isKeyWanted(formattedKey):
                displayData[formattedKey] = _values
        return displayData

    def formatKey(self, key) -> str:
        if self.remove_spaces:
            key = key.replace(" ", "_")
        if not self.keep_case:
            key = key.lower()
        return key

    def isKeyWanted(self, key) -> bool:
        # remove any specifically excluded keys
        if self._keyExclusionfilter is not None and self._keyExclusionfilter.search(
            key
        ):
            # log.debug(f"key_wanted: key {key} matches excl_filter {excl_filter} so key excluded")
            return False
        if self._keyFilter is None:
            # log.debug(
            #    f"key_wanted: No filter and key {key} not excluded by excl_filter {excl_filter} so key wanted"
            # )
            return True
        elif self._keyFilter.search(key):
            # log.debug(
            #    f"key_wanted: key {key} matches filter {filter} and not excl_filter {excl_filter} so key wanted"
            # )
            return True
        else:
            # log.debug(f"key_wanted: key {key} does not match filter {filter} so key excluded")
            return False



import logging
import re

from mppsolar.helpers import get_kwargs, getMaxLen, key_wanted, pad

log = logging.getLogger("table")


class table:
    def output(*args, **kwargs):
        log.info("Using output formatter: table")
        log.debug(f"kwargs {kwargs}")
        data = get_kwargs(kwargs, "data")

        _result = []
        if data is None:
            return _result

        # check if config supplied
        config = get_kwargs(kwargs, "config")
        if config is not None:
            log.debug(f"config: {config}")
            # get formatting info
            remove_spaces = config.get("remove_spaces", True)
            keep_case = config.get("keep_case", False)
            extra_info = config.get("extra_info", False)
            filter = config.get("filter", None)
            excl_filter = config.get("excl_filter", None)

        _filter = None
        _excl_filter = None

        if filter is not None:
            _filter = re.compile(filter)

        if excl_filter is not None:
            _excl_filter = re.compile(excl_filter)

        # remove raw response
        if "raw_response" in data:
            data.pop("raw_response")

        # build header
        if "_command" in data:
            command = data.pop("_command")
        else:
            command = "Unknown command"
        if "_command_description" in data:
            description = data.pop("_command_description")
        else:
            description = "No description found"

        # build data to display
        displayData = {}
        for key in data:
            _values = data[key]
            # remove spaces
            if remove_spaces:
                key = key.replace(" ", "_")
            if not keep_case:
                # make lowercase
                key = key.lower()
            if key_wanted(key, _filter, _excl_filter):
                displayData[key] = _values
        log.debug(f"displayData: {displayData}")

        # build header
        _result.append(f"Command: {command} - {description}")
        if filter or excl_filter:
            _result.append(
                f"Using filter: '{filter}' and excl_filter: '{excl_filter}'. {len(displayData)} results retained from {len(data)} in total"
            )
        _result.append("-" * 80)

        # build data
        maxP = getMaxLen(displayData)
        if maxP < 9:
            maxP = 9
        # maxV = getMaxLen(data.values())
        _result.append(f"{pad('Parameter', maxP+1)}{'Value':<15}\tUnit")
        for key in displayData:
            value = displayData[key][0]
            unit = displayData[key][1]
            if len(displayData[key]) > 2 and displayData[key][2] and extra_info:
                extra = displayData[key][2]
                _result.append(f"{pad(key,maxP+1)}{value:<15}\t{unit:<4}\t{extra}")
            else:
                _result.append(f"{pad(key,maxP+1)}{value:<15}\t{unit:<4}")
        return _result



import logging
import os
import time

from dto.portDTO import PortDTO

from .abstractport import AbstractPort

log = logging.getLogger("USBPort")


class USBPort(AbstractPort):
    def __init__(self, config=None, protocol=None) -> None:
        log.debug(f"Initializing usb port. config:{config}, protocol: {protocol}")
        self.path = config.get("path", None)
        self.protocol = protocol

    def toDTO(self):
        dto = PortDTO(type="usb", path=self.path, protocol=self.protocol.toDTO())
        return dto

    def protocol(self):
        return self.protocol

    def connect(self) -> int:
        log.debug(f"USBPort connecting. path:{self.path}, protocol: {self.protocol}")
        try:
            self.port = os.open(self.path, os.O_RDWR | os.O_NONBLOCK)
            log.debug(f"USBPort port number ${self.port}")
        except Exception as e:
            log.warning(f"Error openning usb port: {e}")
            self.error = e
        return self.port

    def disconnect(self) -> None:
        log.debug(f"USBPort disconnecting {self.port}")
        if self.port is not None:
            os.close(self.port)
        return

    def send_and_receive(self, command) -> dict:
        full_command = self.protocol.get_full_command(command)
        response_line = bytes()

        # Send the command to the open usb connection
        to_send = full_command
        try:
            log.debug(f"length of to_send: {len(to_send)}")
        except:  # noqa: E722
            import pdb

            pdb.set_trace()
        if len(to_send) <= 8:
            # Send all at once
            log.debug("1 chunk send")
            time.sleep(0.35)
            try:
                os.write(self.port, to_send)
            except Exception as e:
                log.debug("USB read error: {}".format(e))

        elif len(to_send) > 8 and len(to_send) < 11:
            log.debug("2 chunk send")
            time.sleep(0.35)
            os.write(self.port, to_send[:5])
            time.sleep(0.35)
            os.write(self.port, to_send[5:])
        else:
            while len(to_send) > 0:
                log.debug("multiple chunk send")
                # Split the byte command into smaller chucks
                send, to_send = to_send[:8], to_send[8:]
                log.debug("send: {}, to_send: {}".format(send, to_send))
                time.sleep(0.35)
                os.write(self.port, send)
        time.sleep(0.25)
        # Read from the usb connection
        # try to a max of 100 times
        for x in range(100):
            # attempt to deal with resource busy and other failures to read
            try:
                time.sleep(0.15)
                r = os.read(self.port, 256)
                response_line += r
            except Exception as e:
                log.debug("USB read error: {}".format(e))
            # Finished is \r is in byte_response
            if bytes([13]) in response_line:
                # remove anything after the \r
                response_line = response_line[: response_line.find(bytes([13])) + 1]
                break
        log.debug("usb response was: %s", response_line)
        return response_line



# Description: Output format for mqtt as individual topics
import logging
from powermon.formats.abstractformat import AbstractFormat

log = logging.getLogger("Topics")


class Topics(AbstractFormat):
    def __init__(self, formatConfig, topic, tag):
        super().__init__(formatConfig)
        self.results_topic = topic
        self.tag = tag

    def sendsMultipleMessages(self) -> bool:
        return True

    def format(self, data):
        log.info("Using output formatter: Topics")
        

        # build topic prefix
        if self.results_topic is not None:
            topic_prefix = self.results_topic + "/" + self.tag
        else:
            topic_prefix = f"{self.tag}/status"


        _data = self.formatAndFilterData(data)

        # Build array of mqtt messages
        msgs = []
        # Loop through responses build topics and messages
        for key in _data:
            value = _data[key][0]
            unit = _data[key][1]
            log.debug(
                f"build_msgs: prefix {topic_prefix}, key {key}, value {value}, unit {unit}"
            )
            msg = {"topic": f"{topic_prefix}/{key}/value", "payload": value}
            msgs.append(msg)
            if unit:
                msg = {"topic": f"{topic_prefix}/{key}/unit", "payload": unit}
                msgs.append(msg)
        log.debug(f"build_msgs: {msgs}")
        return msgs


from abc import ABC, abstractmethod
import logging
from enum import auto
from strenum import LowercaseStrEnum

from mppsolar.protocols.abstractprotocol import AbstractProtocol

# from time import sleep
log = logging.getLogger("Port")


class PortType(LowercaseStrEnum):
    UNKNOWN = auto()
    TEST = auto()
    USB = auto()
    ESP32 = auto()
    SERIAL = auto()
    JKBLE = auto()
    MQTT = auto()
    VSERIAL = auto()
    DALYSERIAL = auto()
    BLE = auto()


class AbstractPort(ABC):
    def __init__(self, config, protocol: AbstractProtocol):
        self.protocol = protocol

    @abstractmethod
    def connect(self) -> None:
        log.debug("Port connect not implemented")
        return

    @abstractmethod
    def disconnect(self) -> None:
        log.debug("Port disconnect not implemented")
        return

    @abstractmethod
    def send_and_receive(self, command) -> dict:
        raise NotImplementedError

    @abstractmethod
    def toDTO(self):
        raise NotImplementedError

    # Question: Should we make this an abstract method?
    def process_command(self, command):
        # Band-aid solution, need to reduce what is sent
        log.debug(f"Command {command}")
        full_command = self.protocol.get_full_command(command)
        log.debug(f"Full Command {full_command}")

        raw_response = self.send_and_receive(command)
        log.debug(f"Send and Receive Response {raw_response}")

        # Handle errors
        # Maybe there should a decode for ERRORs and WARNINGS...
        # Some inverters return the command if the command is unknown:
        if raw_response == full_command:
            return {
                "ERROR": [
                    f"Inverter returned the command string for {command} - the inverter didnt recognise this command",
                    "",
                ]
            }
        # dict is returned on exception
        # Question: What case is this covering?
        if isinstance(raw_response, dict):
            return raw_response

        # Decode response
        decoded_response = self.protocol.decode(raw_response, command)
        log.info(f"Decoded response {decoded_response}")

        return decoded_response



import logging
import random

# from mppsolar.helpers import get_kwargs
from .abstractport import AbstractPort
from dto.portDTO import PortDTO


log = logging.getLogger("test")


class TestPort(AbstractPort):
    def __init__(self, config=None, protocol=None):
        log.debug(f"Initializing test port. config:{config}, protocol: {protocol}")
        self.protocol = protocol

    def __str__(self):
        return "Test port"

    def toDTO(self) -> PortDTO:
        dto = PortDTO(type="test", protocol=self.protocol.toDTO())
        return dto

    def protocol(self):
        return super().protocol()

    def connect(self) -> None:
        log.debug("Test port connected")
        return

    def disconnect(self) -> None:
        log.debug("Test port disconnected")
        return

    def send_and_receive(self, command) -> dict:
        command_defn = self.protocol.get_command_defn(command)

        if command_defn is not None:
            # Have test data defined, so use that
            number_of_test_responses = len(command_defn["test_responses"])
            self._test_data = command_defn["test_responses"][
                random.randrange(number_of_test_responses)
            ]
        else:
            # No test responses defined
            log.warn("Testing a command with no test responses defined")
            self._test_data = None
        response = self._test_data
        log.debug(f"Raw response {response}")
        return response



from pydantic import BaseModel
from .commandScheduleDTO import CommandScheduleDTO
from .deviceDTO import DeviceDTO

class ScheduleDTO(BaseModel):
    loopDuration: int
    device: DeviceDTO
    schedulesCommands: list[CommandScheduleDTO]



import yaml
import logging
import json
from powermon.libs.schedule import Schedule

log = logging.getLogger("APICoordinator")
class ApiCoordinator:
    def __init__(self, config, device, mqtt_broker, schedule : Schedule):
        self.device = device
        self.mqtt_broker = mqtt_broker
        self.schedule = schedule
        self.count = 0
        self.adhocTopic = config.get("adhoc_topic", "powermon/adhoc")
        self.announceTopic = config.get("announce_topic", "powermon/announce")

        self.announceDevice()
        mqtt_broker.subscribe(self.adhocTopic, self.adhocCallback)

        #mqtt_broker.publish(self.announceTopic, self.schedule.getScheduleConfigAsJSON())

    def adhocCallback(self, client, userdata, msg):
        log.info(f"Received `{msg.payload}` on topic `{msg.topic}`")
        yamlString = msg.payload.decode("utf-8")
        log.debug(f"Yaml string: {yamlString}")
        try:
            _command_config = yaml.safe_load(yamlString)
        except yaml.YAMLError as exc:
            log.error(f"Error processing config file: {exc}")

        for command in _command_config["commands"]:
            log.debug(f"command: {command}")
            log.debug(f"self: {self}")
            self.schedule.addOneTimeCommandFromConfig(command)

    def run(self):
        if(self.count > 20):
            log.info("Starting APICoordinator")
            self.announceDevice()
            self.count = 0

        self.count += 1

    def announceDevice(self):
        scheduleDTO = self.schedule.toDTO()
        self.mqtt_broker.publish(self.announceTopic, scheduleDTO.json())


from pydantic import BaseModel
from typing import Optional

from .protocolDTO import ProtocolDTO


class PortDTO(BaseModel):
    type: str
    path: Optional[str]
    baud: Optional[int]
    protocol: ProtocolDTO


from .abstractformat import FormatterType


def getFormatfromConfig(formatConfig, device, topic, tag):
    #Get values from config
    #Type is required
    formatType = formatConfig["type"]

    if formatType == FormatterType.HTMLTABLE:
        from .htmltable import htmltable
        formatter = htmltable(formatConfig)
    elif formatType == FormatterType.HASS:
        from .hass import hass
        formatter = hass(formatConfig, device)
    elif formatType == FormatterType.TOPICS:
        from .topics import Topics
        formatter = Topics(formatConfig, topic, tag)
    elif formatType == FormatterType.SIMPLE:
        from .simple import simple
        formatter = simple(formatConfig)

    return formatter


import logging
from powermon.formats.abstractformat import AbstractFormat

log = logging.getLogger("htmltable")


class htmltable(AbstractFormat):
    def __init__(self, formatConfig):
        super().__init__(formatConfig)
        

    def format(self, data):
        log.info("Using output formatter: htmltable")

        _result = ""
        if data is None:
            return _result
        
        displayData = self.formatAndFilterData(data)
        
        _result += "<table><tr><th>Parameter</th><th>Value</th><th>Unit</th></tr>"
        for key in displayData:
            value = displayData[key][0]
            unit = displayData[key][1]
            _result += f"<tr><td>{key}</td><td>{value}</td><td>{unit}</td></tr>"
        _result += "</table>"
        return _result

        
    
    



from mppsolar.helpers import get_kwargs


class raw:
    def output(*args, **kwargs):
        # print(args, kwargs)
        _result = None
        _data = get_kwargs(kwargs, "data", None)
        if "raw_response" in _data:
            _result = _data["raw_response"][0]
        return _result



import logging
from powermon.formats.abstractformat import AbstractFormat

log = logging.getLogger("simple")


class simple(AbstractFormat):
    def __init__(self, formatConfig):
        super().__init__(formatConfig)
        self.extra_info = formatConfig.get("extra_info", False)

    def format(self, data):
        log.info("Using output formatter: simple")

        _result = []
        if data is None:
            return _result

        displayData = self.formatAndFilterData(data)
        log.debug(f"displayData: {displayData}")

        # build data to display
        for key in displayData:
            value = displayData[key][0]
            unit = displayData[key][1]
            if len(displayData[key]) > 2 and displayData[key][2] and self.extra_info:
                extra = displayData[key][2]
                _result.append(f"{key}={value}{unit} {extra}")
            else:
                _result.append(f"{key}={value}{unit}")
        return _result



import logging
import re

from mppsolar.helpers import get_kwargs
from powermon.outputs.abstractoutput import AbstractOutput

log = logging.getLogger("MQTT")


class MQTT(AbstractOutput):
    
    def __init__(self, output_config, mqtt_broker, formatter) -> None:
        super().__init__(formatter)
        self.mqtt_broker = mqtt_broker
        self.results_topic = output_config.get("topic", None)
        self.tag = output_config.get("tag", None)

        if self.tag is None:
            raise Exception("Configuration Error. Output type MQTT requires a tag to be set")

        
    
    def __str__(self):
        return "outputs the results to the supplied mqtt broker: eg {tag}/status/total_output_active_power/value 1250"

    

    def output(self, data):
        log.info("Using output processor: mqtt")
        # exit if no data
        if data is None:
            return

        # exit if no broker
        if self.mqtt_broker is None:
            log.error("No mqtt broker supplied")
            return
        
        # build topic prefix
        if self.results_topic is not None:
            self.topic_prefix = self.results_topic + "/" + self.tag
        else:
            #TODO: if tag is null get topic from data
            self.topic_prefix = f"{self.tag}/status"

        # build the messages...
        formattedData = self.formatter.format(data)
        log.debug(f"mqtt.output msgs {formattedData}")

        # publish
        if (self.formatter.sendsMultipleMessages()):
            self.mqtt_broker.publishMultiple(formattedData)
        else:
            self.mqtt_broker.publish(self.topic_prefix, formattedData)


import logging

from mppsolar.helpers import get_kwargs
from powermon.outputs.abstractoutput import AbstractOutput

log = logging.getLogger("screen")


class Screen(AbstractOutput):
    def __init__(self, outputConfig, formatter):
        super().__init__(formatter)
        self.formatter = formatter

    def output(self, data):
        log.info("Using output sender: screen")
        if data is None:
            return

        formatted_data = self.formatter.format(data)
        if formatted_data is None:
            print("Nothing returned from data formatting")
            return

        if isinstance(formatted_data, list):
            for line in formatted_data:
                print(line)
        else:
            print(formatted_data)



from powermon.formats import getFormatfromConfig
from .abstractoutput import OutputType

def getOutputFromConfig(outputConfig, device, mqtt_broker):
    outputType = outputConfig["type"]

    formatConfig = outputConfig["format"]
    topic = outputConfig.get("topic", None)
    tag = outputConfig.get("tag", None)
    format = getFormatfromConfig(formatConfig, device, topic, tag)

    output_class = None
    #Only import the required class
    if outputType == OutputType.SCREEN:
        from .screen import Screen
        output_class = Screen(outputConfig, format)
    elif outputType == OutputType.MQTT:
        from .mqtt import MQTT
        output_class = MQTT(outputConfig, mqtt_broker, format)
        

    return output_class


import logging
import serial
import time
from dto.portDTO import PortDTO

from .abstractport import AbstractPort

log = logging.getLogger("SerialPort")


class SerialPort(AbstractPort):
    def __init__(self, config=None, protocol=None) -> None:
        log.debug(f"Initializing usbserial port. config:{config}, protocol: {protocol}")
        self.path = config.get("path", None)
        self.baud = config.get("baud", None)
        self.protocol = protocol
        self.serialPort = None
        self.error = None

    def toDTO(self) -> PortDTO:
        dto = PortDTO(type="serial", path=self.path, baud=self.baud, protocol=self.protocol.toDTO())
        return dto

    def connect(self) -> None:
        log.debug(f"usbserial port connecting. path:{self.path}, baud:{self.baud}")
        try:
            self.serialPort = serial.Serial(port=self.path, baudrate=self.baud, timeout=1, write_timeout=1)
        except Exception as e:
            log.error(f"Error openning serial port: {e}")
            self.error = e
        return

    def disconnect(self) -> None:
        log.debug("usbserial port disconnecting")
        if self.serialPort is not None:
            self.serialPort.close()
        return

    def send_and_receive(self, command) -> dict:
        full_command = self.protocol.get_full_command(command)
        response_line = None
        log.debug(f"port {self.serialPort}")
        if self.serialPort is None:
            log.error("Port not available")
            return {"ERROR": [f"Serial command execution failed {self.error}", ""]}
        try:
            log.debug("Executing command via usbserial...")
            self.serialPort.reset_input_buffer()
            self.serialPort.reset_output_buffer()
            c = self.serialPort.write(full_command)
            log.debug(f"Wrote {c} bytes")
            self.serialPort.flush()
            time.sleep(0.1)  # give serial port time to receive the data
            response_line = self.serialPort.read_until(b"\r")
            log.debug("serial response was: %s", response_line)
            return response_line
        except Exception as e:
            log.warning(f"Serial read error: {e}")
        log.info("Command execution failed")
        return {"ERROR": ["Serial command execution failed", ""]}



import logging

from mppsolar.protocols import get_protocol
from powermon.ports.abstractport import PortType
from powermon.ports.serialport import SerialPort
from powermon.ports.usbport import USBPort

# Set-up logger
log = logging.getLogger("ports")


def getPortFromConfig(port_config):
    log.debug(f"port_config: {port_config}")

    portObject = None

    if not port_config:
        log.info("no port config supplied, defaulting to test port")
        port_config = {"type": "test", "protocol": "PI30"}

    # get protocol handler
    protocol = get_protocol(protocol=port_config.get("protocol"))

    # port type is mandatory
    portType = port_config.get("type")
    log.debug(f"portType: {portType}")

    # return None if port type is not defined
    if portType is None:
        return None

    # build port object
    if portType == PortType.SERIAL:
        portObject = SerialPort(config=port_config, protocol=protocol)
    elif portType == PortType.USB:
        portObject = USBPort(config=port_config, protocol=protocol)

    # Pattern for port types that cause problems when imported
    elif portType == PortType.TEST:
        log.debug("portType test found")
        from powermon.ports.testport import TestPort

        portObject = TestPort(config=port_config, protocol=protocol)

    else:
        log.info("port type object not found for %s", portType)

    return portObject
